@@ -20,7 +20,10 @@
 Utilities with minimum-depends for use in setup.py
 """
 
+from __future__ import unicode_literals
+
 import email
+import io
 import os
 import re
 import subprocess
@@ -37,13 +40,12 @@ from setuptools.command import install_scripts
 from setuptools.command import sdist
 
 try:
-    import cStringIO as io
+    import cStringIO
 except ImportError:
-    import io
+    import io as cStringIO
 
 from pbr import extra_files
 
-log.set_verbosity(log.INFO)
 TRUE_VALUES = ('true', '1', 'yes')
 REQUIREMENTS_FILES = ('requirements.txt', 'tools/pip-requires')
 TEST_REQUIREMENTS_FILES = ('test-requirements.txt', 'tools/test-requires')
@@ -56,7 +58,13 @@ def get_requirements_files():
     files = os.environ.get("PBR_REQUIREMENTS_FILES")
     if files:
         return tuple(f.strip() for f in files.split(','))
-    return REQUIREMENTS_FILES
+    # Returns a list composed of:
+    # - REQUIREMENTS_FILES with -py2 or -py3 in the name
+    #   (e.g. requirements-py3.txt)
+    # - REQUIREMENTS_FILES
+    return (list(map(('-py' + str(sys.version_info[0])).join,
+                     map(os.path.splitext, REQUIREMENTS_FILES)))
+            + list(REQUIREMENTS_FILES))
 
 
 def append_text_list(config, key, text_list):
@@ -69,18 +77,6 @@ def append_text_list(config, key, text_list):
     config[key] = '\n'.join(new_value)
 
 
-def _parse_mailmap(mailmap_info):
-    mapping = dict()
-    for l in mailmap_info:
-        try:
-            canonical_email, alias = re.match(
-                r'[^#]*?(<.+>).*(<.+>).*', l).groups()
-        except AttributeError:
-            continue
-        mapping[alias] = canonical_email
-    return mapping
-
-
 def _pip_install(links, requires, root=None, option_dict=dict()):
     if get_boolean_option(
             option_dict, 'skip_pip_install', 'SKIP_PIP_INSTALL'):
@@ -91,30 +87,11 @@ def _pip_install(links, requires, root=None, option_dict=dict()):
     for link in links:
         cmd.append("-f")
         cmd.append(link)
+
+    # NOTE(ociuhandu): popen on Windows does not accept unicode strings
     _run_shell_command(
         cmd + requires,
-        throw_on_error=True, buffer=False)
-
-
-def read_git_mailmap(root_dir=None, mailmap='.mailmap'):
-    if not root_dir:
-        root_dir = _run_shell_command(
-            ['git', 'rev-parse', '--show-toplevel'])
-
-    mailmap = os.path.join(root_dir, mailmap)
-    if os.path.exists(mailmap):
-        return _parse_mailmap(open(mailmap, 'r').readlines())
-
-    return dict()
-
-
-def canonicalize_emails(changelog, mapping):
-    """Takes in a string and an email alias mapping and replaces all
-       instances of the aliases in the string with their real email.
-    """
-    for alias, email_address in mapping.items():
-        changelog = changelog.replace(alias, email_address)
-    return changelog
+        throw_on_error=True, buffer=False, env=dict(PIP_USE_WHEEL=b"true"))
 
 
 def _any_existing(file_list):
@@ -148,29 +125,37 @@ def parse_requirements(requirements_files=None):
         if (not line.strip()) or line.startswith('#'):
             continue
 
+        try:
+            project_name = pkg_resources.Requirement.parse(line).project_name
+        except ValueError:
+            project_name = None
+
         # For the requirements list, we need to inject only the portion
         # after egg= so that distutils knows the package it's looking for
         # such as:
         # -e git://github.com/openstack/nova/master#egg=nova
         # -e git://github.com/openstack/nova/master#egg=nova-1.2.3
         if re.match(r'\s*-e\s+', line):
-            requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$',
-                                       egg_fragment,
-                                       line))
+            line = re.sub(r'\s*-e\s+.*#egg=(.*)$', egg_fragment, line)
         # such as:
         # http://github.com/openstack/nova/zipball/master#egg=nova
         # http://github.com/openstack/nova/zipball/master#egg=nova-1.2.3
         elif re.match(r'\s*https?:', line):
-            requirements.append(re.sub(r'\s*https?:.*#egg=(.*)$',
-                                       egg_fragment,
-                                       line))
+            line = re.sub(r'\s*https?:.*#egg=(.*)$', egg_fragment, line)
         # -f lines are for index locations, and don't get used here
         elif re.match(r'\s*-f\s+', line):
-            pass
-        elif line in BROKEN_ON_27 and sys.version_info >= (2, 7):
-            pass
-        else:
+            line = None
+            reason = 'Index Location'
+        elif (project_name and
+                project_name in BROKEN_ON_27 and sys.version_info >= (2, 7)):
+            line = None
+            reason = 'Python 2.6 only dependency'
+
+        if line is not None:
             requirements.append(line)
+        else:
+            log.info(
+                '[pbr] Excluding %s: %s' % (project_name, reason))
 
     return requirements
 
@@ -201,7 +186,7 @@ def _run_git_command(cmd, git_dir, **kwargs):
         ['git', '--git-dir=%s' % git_dir] + cmd, **kwargs)
 
 
-def _run_shell_command(cmd, throw_on_error=False, buffer=True):
+def _run_shell_command(cmd, throw_on_error=False, buffer=True, env=None):
     if buffer:
         out_location = subprocess.PIPE
         err_location = subprocess.PIPE
@@ -209,9 +194,14 @@ def _run_shell_command(cmd, throw_on_error=False, buffer=True):
         out_location = None
         err_location = None
 
+    newenv = os.environ.copy()
+    if env:
+        newenv.update(env)
+
     output = subprocess.Popen(cmd,
                               stdout=out_location,
-                              stderr=err_location)
+                              stderr=err_location,
+                              env=newenv)
     out = output.communicate()
     if output.returncode and throw_on_error:
         raise distutils.errors.DistutilsError(
@@ -226,7 +216,23 @@ def _get_git_directory():
 
 
 def _git_is_installed():
-    return _run_shell_command(['which', 'git'])
+    try:
+        # We cannot use 'which git' as it may not be available
+        # in some distributions, So just try 'git --version'
+        # to see if we run into trouble
+        _run_shell_command(['git', '--version'])
+    except OSError:
+        return False
+    return True
+
+
+def _get_highest_tag(tags):
+    """Find the highest tag from a list.
+
+    Pass in a list of tag strings and this will return the highest
+    (latest) as sorted by the pkg_resources version parser.
+    """
+    return max(tags, key=pkg_resources.parse_version)
 
 
 def get_boolean_option(option_dict, option_name, env_name):
@@ -250,11 +256,44 @@ def write_git_changelog(git_dir=None, dest_dir=os.path.curdir,
         if git_dir is None:
             git_dir = _get_git_directory()
         if git_dir:
-            changelog = _run_git_command('log', git_dir)
-            mailmap = read_git_mailmap()
-            with open(new_changelog, "wb") as changelog_file:
-                changelog_file.write(canonicalize_emails(
-                    changelog, mailmap).encode('utf-8'))
+            log_cmd = ['log', '--oneline', '--decorate']
+            changelog = _run_git_command(log_cmd, git_dir)
+            first_line = True
+            with io.open(new_changelog, "w",
+                         encoding="utf-8") as changelog_file:
+                changelog_file.write("CHANGES\n=======\n\n")
+                for line in changelog.split('\n'):
+                    line_parts = line.split()
+                    if len(line_parts) < 2:
+                        continue
+                    # Tags are in a list contained in ()'s. If a commit
+                    # subject that is tagged happens to have ()'s in it
+                    # this will fail
+                    if line_parts[1].startswith('(') and ')' in line:
+                        msg = line.split(')')[1].strip()
+                    else:
+                        msg = " ".join(line_parts[1:])
+
+                    if "tag:" in line:
+                        tags = [
+                            tag.split(",")[0]
+                            for tag in line.split(")")[0].split("tag: ")[1:]]
+                        tag = _get_highest_tag(tags)
+
+                        underline = len(tag) * '-'
+                        if not first_line:
+                            changelog_file.write('\n')
+                        changelog_file.write(
+                            ("%(tag)s\n%(underline)s\n\n" %
+                             dict(tag=tag,
+                                  underline=underline)))
+
+                    if not msg.startswith("Merge "):
+                        if msg.endswith("."):
+                            msg = msg[:-1]
+                        changelog_file.write(
+                            ("* %(msg)s\n" % dict(msg=msg)))
+                    first_line = False
 
 
 def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
@@ -269,14 +308,14 @@ def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
                 and not os.access(new_authors, os.W_OK)):
             return
         log.info('[pbr] Generating AUTHORS')
-        ignore_emails = '(jenkins@review|infra@lists)'
+        ignore_emails = '(jenkins@review|infra@lists|jenkins@openstack)'
         if git_dir is None:
             git_dir = _get_git_directory()
         if git_dir:
             authors = []
 
             # don't include jenkins email address in AUTHORS file
-            git_log_cmd = ['log', '--format=%aN <%aE>']
+            git_log_cmd = ['log', '--use-mailmap', '--format=%aN <%aE>']
             authors += _run_git_command(git_log_cmd, git_dir).split('\n')
             authors = [a for a in authors if not re.search(ignore_emails, a)]
 
@@ -288,11 +327,6 @@ def generate_authors(git_dir=None, dest_dir='.', option_dict=dict()):
                           for signed in co_authors if signed]
 
             authors += co_authors
-
-            # canonicalize emails, remove duplicates and sort
-            mailmap = read_git_mailmap(git_dir)
-            authors = canonicalize_emails('\n'.join(authors), mailmap)
-            authors = authors.split('\n')
             authors = sorted(set(authors))
 
             with open(new_authors, 'wb') as new_authors_fh:
@@ -651,7 +685,7 @@ try:
 
         def _sphinx_run(self):
             if not self.verbose:
-                status_stream = io.StringIO()
+                status_stream = cStringIO.StringIO()
             else:
                 status_stream = sys.stdout
             confoverrides = {}
@@ -662,6 +696,7 @@ try:
             if self.today:
                 confoverrides['today'] = self.today
             sphinx_config = config.Config(self.config_dir, 'conf.py', {}, [])
+            sphinx_config.init_values()
             if self.builder == 'man' and len(sphinx_config.man_pages) == 0:
                 return
             app = application.Sphinx(
@@ -714,6 +749,13 @@ try:
                 else:
                     setup_command.BuildDoc.run(self)
 
+        def finalize_options(self):
+            # Not a new style class, super keyword does not work.
+            setup_command.BuildDoc.finalize_options(self)
+            # Allow builders to be configurable - as a comma separated list.
+            if not isinstance(self.builders, list) and self.builders:
+                self.builders = self.builders.split(',')
+
     class LocalBuildLatex(LocalBuildDoc):
         builders = ['latex']
         command_name = 'build_sphinx_latex'
